@@ -3,35 +3,38 @@
 import asyncio.subprocess
 import hashlib
 import os
-import pathlib
 import shutil
 import tempfile
-from typing import List, Optional, Tuple, Sequence
+import urllib.parse
+import zipfile
+from typing import List, Optional, Sequence, Tuple, Union
 
-from meadowrun._vendor.aiodocker import exceptions as aiodocker_exceptions
 import filelock
 
-from meadowrun.aws_integration.aws_core import _get_default_region_name
+from meadowrun._vendor.aiodocker import exceptions as aiodocker_exceptions
+from meadowrun.aws_integration import s3
 from meadowrun.aws_integration.ecr import (
-    does_image_exist,
-    ensure_repository,
-    get_username_password,
+    get_ecr_helper,
 )
 from meadowrun.aws_integration.management_lambdas.ec2_alloc_stub import (
     _MEADOWRUN_GENERATED_DOCKER_REPO,
 )
+from meadowrun.azure_integration.acr import get_acr_helper
 from meadowrun.credentials import RawCredentials, SshKey
 from meadowrun.docker_controller import (
-    build_image,
-    push_image,
-    pull_image,
     _does_digest_exist_locally,
+    build_image,
+    pull_image,
+    push_image,
 )
 from meadowrun.meadowrun_pb2 import (
+    EnvironmentSpec,
     EnvironmentSpecInCode,
+    EnvironmentType,
     Job,
     ServerAvailableContainer,
 )
+from meadowrun.run_job_core import CloudProviderType, ContainerRegistryHelper
 
 _GIT_REPO_URL_SUFFIXES_TO_REMOVE = [".git", "/"]
 
@@ -102,11 +105,12 @@ async def get_code_paths(
     local_copies_folder: str,
     job: Job,
     credentials: Optional[RawCredentials],
-) -> Sequence[str]:
+) -> Tuple[Sequence[str], Optional[str]]:
     """
-    Returns code_paths based on Job.code_deployment. code_paths will have paths to
-    folders that are available on this machine that contain "the user's code". Code
-    paths can be empty, which would mean that we don't need any user code.
+    Returns [code_paths, interpreter_spec_path] based on Job.code_deployment. code_paths
+    will have paths to folders that are available on this machine that contain "the
+    user's code". interpreter_spec_path will be a path where we can find e.g. an
+    environment.yml file that can be used to build a conda environment.
 
     For ServerAvailableFolder, this function will effectively use code_paths as
     specified.
@@ -116,14 +120,23 @@ async def get_code_paths(
     """
     case = job.WhichOneof("code_deployment")
     if case == "server_available_folder":
-        return job.server_available_folder.code_paths
+        if job.server_available_folder.code_paths:
+            interpreter_spec_path = job.server_available_folder.code_paths[0]
+        else:
+            interpreter_spec_path = None
+        return (
+            job.server_available_folder.code_paths,
+            # TODO theoretically the environment file could be not in the first code
+            # path, but should be a very unlikely scenario for now.
+            interpreter_spec_path,
+        )
     elif case == "git_repo_commit":
         return await _get_git_code_paths(
             git_repos_folder,
             local_copies_folder,
             job.git_repo_commit.repo_url,
             job.git_repo_commit.commit,
-            job.git_repo_commit.path_in_repo,
+            job.git_repo_commit.path_to_source,
             credentials,
         )
     elif case == "git_repo_branch":
@@ -136,9 +149,16 @@ async def get_code_paths(
             # TODO this is a bit of a hack. We don't want to bother pulling/merging e.g.
             # origin/main into main, so we just always reference origin/main.
             f"origin/{job.git_repo_branch.branch}",
-            job.git_repo_branch.path_in_repo,
+            job.git_repo_branch.path_to_source,
             credentials,
         )
+    elif case == "code_zip_file":
+        return (
+            await _get_zip_file_code_paths(
+                local_copies_folder, job.code_zip_file.url, job.code_zip_file.code_paths
+            ),
+            None,
+        )
     else:
         raise ValueError(f"Unrecognized code_deployment {case}")
 
@@ -167,9 +187,9 @@ async def _get_git_code_paths(
     local_copies_folder: str,
     repo_url: str,
     revision_spec: str,
-    path_in_repo: str,
+    path_to_source: str,
     credentials: Optional[RawCredentials],
-) -> Sequence[str]:
+) -> Tuple[Sequence[str], str]:
     """Returns code_paths for GitRepoCommit"""
 
     local_clone_name = _get_git_repo_local_clone_name(repo_url)
@@ -213,14 +233,42 @@ async def _get_git_code_paths(
             shutil.copytree(local_path, local_copy_path)
 
         # TODO raise a friendlier exception if this path doesn't exist
-        path_in_local_copy = str(
-            (pathlib.Path(local_copy_path) / path_in_repo).resolve()
-        )
-        return [path_in_local_copy]
+        return [os.path.join(local_copy_path, path_to_source)], local_copy_path
+
+
+async def _get_zip_file_code_paths(
+    local_copies_folder: str, zip_file_url: str, code_paths: Sequence[str]
+) -> List[str]:
+    decoded_url = urllib.parse.urlparse(zip_file_url)
+    if decoded_url.scheme == "file":
+        with zipfile.ZipFile(decoded_url.path) as zip_file:
+            extracted_folder = os.path.splitext(os.path.basename(decoded_url.path))[0]
+            zip_file.extractall(os.path.join(local_copies_folder, extracted_folder))
+        return [
+            os.path.join(local_copies_folder, extracted_folder, zip_path)
+            for zip_path in code_paths
+        ]
+
+    if decoded_url.scheme == "s3":
+        bucket_name = decoded_url.netloc
+        object_name = decoded_url.path.lstrip("/")
+        extracted_folder = os.path.join(local_copies_folder, object_name)
+
+        if not os.path.exists(extracted_folder):
+            zip_file_path = extracted_folder + ".zip"
+            await s3.download_file(bucket_name, object_name, zip_file_path)
+            with zipfile.ZipFile(zip_file_path) as zip_file:
+                zip_file.extractall(os.path.join(local_copies_folder, extracted_folder))
+
+        return [os.path.join(extracted_folder, zip_path) for zip_path in code_paths]
+
+    raise ValueError(f"Unknown URL scheme in {zip_file_url}")
 
 
 async def compile_environment_spec_to_container(
-    environment_spec_in_code: EnvironmentSpecInCode, code_paths: Sequence[str]
+    environment_spec: Union[EnvironmentSpecInCode, EnvironmentSpec],
+    interpreter_spec_path: str,
+    cloud: Optional[Tuple[CloudProviderType, str]],
 ) -> ServerAvailableContainer:
     """
     Turns e.g. a conda_environment.yml file into a docker container which will be
@@ -230,48 +278,47 @@ async def compile_environment_spec_to_container(
     TODO we should also consider creating conda environments locally rather than always
     making a container for them, that might be more efficient.
     """
-    if (
-        environment_spec_in_code.environment_type
-        == EnvironmentSpecInCode.EnvironmentType.CONDA
-    ):
-        # TODO theoretically the environment file could be not in the first code path,
-        # but should be a very unlikely scenario for now.
-        path_to_spec = os.path.join(
-            code_paths[0], environment_spec_in_code.path_to_spec
+    if environment_spec.environment_type == EnvironmentType.CONDA:
+        path_to_spec, spec_hash = _get_path_and_hash(
+            environment_spec, interpreter_spec_path
         )
-        with open(path_to_spec, "rb") as spec:
-            # TODO probably better to exclude the name and prefix in the file as those
-            # are ignored
-            spec_hash = hashlib.blake2b(spec.read(), digest_size=64).hexdigest()
 
-        region_name = await _get_default_region_name()
-        repository_prefix = ensure_repository(
-            _MEADOWRUN_GENERATED_DOCKER_REPO, region_name
-        )
-        image_name = (
-            f"{repository_prefix}/{_MEADOWRUN_GENERATED_DOCKER_REPO}:{spec_hash}"
-        )
-        result = ServerAvailableContainer(image_name=image_name)
+        if cloud is None:
+            helper = ContainerRegistryHelper(
+                False, None, f"{_MEADOWRUN_GENERATED_DOCKER_REPO}:{spec_hash}", False
+            )
+        else:
+            cloud_type, region_name = cloud
+            if cloud_type == "EC2":
+                helper = await get_ecr_helper(
+                    _MEADOWRUN_GENERATED_DOCKER_REPO, spec_hash, region_name
+                )
+            elif cloud_type == "AzureVM":
+                helper = await get_acr_helper(
+                    _MEADOWRUN_GENERATED_DOCKER_REPO, spec_hash, region_name
+                )
+            else:
+                raise ValueError(f"Unexpected cloud_type {cloud_type}")
 
-        username_password = get_username_password(region_name)
+        result = ServerAvailableContainer(image_name=helper.image_name)
 
         # if the image already exists locally, just return it
-        if await _does_digest_exist_locally(image_name):
+        if await _does_digest_exist_locally(helper.image_name):
             return result
 
         # if the image doesn't exist locally but does exist in ECR, try to pull it
-        if does_image_exist(_MEADOWRUN_GENERATED_DOCKER_REPO, spec_hash, region_name):
+        if helper.does_image_exist:
             try:
-                await pull_image(image_name, username_password)
+                await pull_image(helper.image_name, helper.username_password)
                 # TODO we're assuming that once we pull the image it will be available
                 # locally until the job runs, which might not be true if we implement
                 # something to clean unused images.
                 return result
             except (aiodocker_exceptions.DockerError, ValueError) as e:
                 print(
-                    f"Warning, image {image_name} was supposed to exist, but we were "
-                    "unable to pull, so re-building locally. This could happen if the "
-                    "image was deleted between when we checked for it and when we "
+                    f"Warning, image {helper.image_name} was supposed to exist, but we "
+                    "were unable to pull, so re-building locally. This could happen if "
+                    "the image was deleted between when we checked for it and when we "
                     f"pulled, but that should be rare: {e}"
                 )
 
@@ -283,7 +330,7 @@ async def compile_environment_spec_to_container(
         )
         await build_image(
             [(docker_file_path, "Dockerfile"), (path_to_spec, spec_filename)],
-            image_name,
+            helper.image_name,
             {"ENV_FILE": spec_filename},
         )
 
@@ -292,16 +339,48 @@ async def compile_environment_spec_to_container(
         # isn't critical.
         # TODO we should also somehow avoid multiple processes on the same (or even on
         # different machines) building the identical image at the same time.
-        try:
-            await push_image(image_name, username_password)
-        except aiodocker_exceptions.DockerError as e:
-            print(
-                f"Warning, image {image_name} couldn't be pushed. Execution can "
-                f"continue, but we won't be able to reuse this image later: {e}"
-            )
+        if helper.should_push:
+            try:
+                await push_image(helper.image_name, helper.username_password)
+            except aiodocker_exceptions.DockerError as e:
+                print(
+                    f"Warning, image {helper.image_name} couldn't be pushed. Execution "
+                    f"can continue, but we won't be able to reuse this image later: {e}"
+                )
 
         return result
     else:
         raise ValueError(
-            f"Unexpected environment_type {environment_spec_in_code.environment_type}"
+            f"Unexpected environment_type {environment_spec.environment_type}"
         )
+
+
+def _get_path_and_hash(
+    environment_spec: Union[EnvironmentSpecInCode, EnvironmentSpec],
+    interpreter_spec_path: str,
+) -> Tuple[str, str]:
+    def hash_spec(it: bytes) -> str:
+        # TODO probably better to exclude the name and prefix in the file as those
+        # are ignored
+        return hashlib.blake2b(it, digest_size=64).hexdigest()
+
+    if isinstance(environment_spec, EnvironmentSpecInCode):
+        # in this case, interpreter_spec_path is a path to the root of the place where
+        # the spec lives.
+        path_to_spec = os.path.join(
+            interpreter_spec_path, environment_spec.path_to_spec
+        )
+        with open(path_to_spec, "rb") as spec:
+            spec_hash = hash_spec(spec.read())
+        return path_to_spec, spec_hash
+    elif isinstance(environment_spec, EnvironmentSpec):
+        # in this case, interpreter_spec_path is a path to where we save the spec for
+        # later processing.
+        spec_hash = hash_spec(environment_spec.spec.encode("UTF-8"))
+        path_to_spec = os.path.join(
+            interpreter_spec_path, f"conda_env_spec_{spec_hash}.yml"
+        )
+        with open(path_to_spec, "w") as env_spec:
+            env_spec.write(environment_spec.spec)
+
+        return path_to_spec, spec_hash
