@@ -1,20 +1,20 @@
 from __future__ import annotations
 
-import abc
 import asyncio
 import dataclasses
-import functools
-import io
 import os.path
 import pickle
+import platform
 import shlex
-import threading
+import shutil
+import tempfile
+import urllib.parse
 import uuid
+from enum import Enum
 from typing import (
     Any,
     Callable,
     Dict,
-    Generic,
     Iterable,
     List,
     Optional,
@@ -26,13 +26,23 @@ from typing import (
 )
 
 import cloudpickle
-import fabric
-import paramiko.ssh_exception
 
-from meadowrun.aws_integration.ssh_keys import ensure_meadowrun_key_pair
-from meadowrun.run_job_local import run_local
-from meadowrun.aws_integration.aws_core import _get_default_region_name
-from meadowrun.config import MEADOWRUN_INTERPRETER, JOB_ID_VALID_CHARACTERS
+from meadowrun import local_code
+from meadowrun.aws_integration import s3
+from meadowrun.aws_integration.ec2_instance_allocation import (
+    run_job_ec2_instance_registrar,
+)
+from meadowrun.aws_integration.grid_tasks_sqs import prepare_ec2_run_map
+from meadowrun.azure_integration.azure_instance_allocation import (
+    run_job_azure_vm_instance_registrar,
+)
+from meadowrun.azure_integration.azure_ssh_keys import get_meadowrun_vault_name
+from meadowrun.azure_integration.grid_tasks_queue import prepare_azure_vm_run_map
+from meadowrun.azure_integration.mgmt_functions.azure.azure_rest_api import (
+    get_subscription_id_sync,
+)
+from meadowrun.conda import env_export
+from meadowrun.config import JOB_ID_VALID_CHARACTERS, MEADOWRUN_INTERPRETER
 from meadowrun.credentials import CredentialsSourceForService
 from meadowrun.deployment import (
     CodeDeployment,
@@ -40,23 +50,20 @@ from meadowrun.deployment import (
     VersionedCodeDeployment,
     VersionedInterpreterDeployment,
 )
-from meadowrun.aws_integration.ec2_alloc import allocate_ec2_instances
-from meadowrun.aws_integration.grid_tasks_sqs import (
-    get_results,
-    worker_loop,
-    create_queues_and_add_tasks,
-)
 from meadowrun.meadowrun_pb2 import (
     AwsSecret,
+    AzureSecret,
+    CodeZipFile,
     ContainerAtDigest,
     ContainerAtTag,
     Credentials,
     CredentialsSourceMessage,
+    EnvironmentSpec,
     EnvironmentSpecInCode,
+    EnvironmentType,
     GitRepoBranch,
     GitRepoCommit,
     Job,
-    ProcessState,
     PyCommandJob,
     PyFunctionJob,
     QualifiedFunctionName,
@@ -66,7 +73,7 @@ from meadowrun.meadowrun_pb2 import (
     ServerAvailableInterpreter,
     StringPair,
 )
-from meadowrun.instance_selection import Resources
+from meadowrun.run_job_core import CloudProviderType, Host, JobCompletion, SshHost
 
 _T = TypeVar("_T")
 _U = TypeVar("_U")
@@ -77,25 +84,6 @@ _U = TypeVar("_U")
 _DEFAULT_CONCURRENT_TASKS_FACTOR = 0.5
 
 
-async def _retry(
-    function: Callable[[], _T],
-    exception_types: Union[Exception, Tuple[Exception, ...]],
-    max_num_attempts: int = 5,
-    delay_seconds: float = 1,
-) -> _T:
-    i = 0
-    while True:
-        try:
-            return function()
-        except exception_types as e:  # type: ignore
-            i += 1
-            if i >= max_num_attempts:
-                raise
-            else:
-                print(f"Retrying on error: {e}")
-                await asyncio.sleep(delay_seconds)
-
-
 @dataclasses.dataclass(frozen=True)
 class Deployment:
     interpreter: Union[
@@ -103,7 +91,51 @@ class Deployment:
     ] = None
     code: Union[CodeDeployment, VersionedCodeDeployment, None] = None
     environment_variables: Optional[Dict[str, str]] = None
-    credentials_sources: Optional[List[CredentialsSourceForService]] = None
+    credentials_sources: List[CredentialsSourceForService] = dataclasses.field(
+        default_factory=list
+    )
+
+    @classmethod
+    async def mirror_local(
+        cls,
+        include_sys_path: bool = True,
+        additional_paths: Iterable[str] = tuple(),
+        conda_env: Optional[str] = None,
+        environment_variables: Optional[Dict[str, str]] = None,
+    ) -> Deployment:
+        """A deployment that mirrors local environment and code.
+
+        :param include_sys_path: if True, find paths on sys.path, copy them to
+            the remote machines, and add them to sys.path there. Ignores any installed
+            packages. Defaults to True.
+        :param additional_paths: local code paths to mirror on the remote servers.
+        :param conda_env: name or full path of locally installed conda environment to
+            use. Defaults to None which means the currently activated env.
+        :param environment_variables: e.g. :code:`{"PYTHONHASHSEED": "0"}`. These
+            environment variables will be set in the remote environment.
+        """
+        if platform.system() != "Linux":
+            raise ValueError(
+                "mirror_local is only supported on Linux "
+                "because Conda environments are not cross-platform."
+            )
+        interpreter = EnvironmentSpec(
+            environment_type=EnvironmentType.CONDA,
+            spec=await env_export(conda_env),
+        )
+
+        code: CodeDeployment = ServerAvailableFolder(code_paths=[])
+        # annoyingly, this tmp dir now gets deleted in run_local when the file
+        # has been uploaded/unpacked depending on the Host implementation
+        tmp_dir = tempfile.mkdtemp()
+        zip_file_path, zip_paths = local_code.zip(
+            tmp_dir, include_sys_path, additional_paths
+        )
+
+        url = urllib.parse.urlunparse(("file", "", zip_file_path, "", "", ""))
+        code = CodeZipFile(url=url, code_paths=zip_paths)
+
+        return cls(interpreter, code, environment_variables, [])
 
     @classmethod
     def git_repo(
@@ -111,24 +143,40 @@ class Deployment:
         repo_url: str,
         branch: Optional[str] = None,
         commit: Optional[str] = None,
-        path_in_repo: Optional[str] = None,
+        path_to_source: Optional[str] = None,
         conda_yml_file: Optional[str] = None,
         environment_variables: Optional[Dict[str, str]] = None,
         ssh_key_aws_secret: Optional[str] = None,
+        ssh_key_azure_secret: Optional[str] = None,
     ) -> Deployment:
         """
-        A deployment based on a git repo:
-        - repo_url: e.g. https://github.com/meadowdata/test_repo
-        - branch: defaults to "main" if neither branch nor commit are specified.
-        - commit: can be provided instead of branch to use a specific commit hash
-        - path_in_repo: e.g. "src/python" to use a subdirectory of the repo
-        - conda_yml_file: a directory (relative to the repo, or path_in_repo if that's
-          specified) that points to a file generated by `conda env export`. This file
-          will be used to generate the environment to run in.
-        - environment_variables: see class variable
-        - ssh_key_aws_secret: The name of an AWS secret that has the
+        A deployment based on a git repo.
+
+        Args:
+            repo_url: e.g. `"https://github.com/meadowdata/test_repo"`
+            branch: defaults to `"main"` if neither branch nor commit are specified.
+            commit: can be provided instead of branch to use a specific commit hash,
+                e.g. `"d018b54"`
+            path_to_source: e.g. `"src/python"` to use a subdirectory of the repo
+            conda_yml_file: a file (relative to the repo, note that this IGNORES
+                `path_to_source`) file generated by `conda env export`, e.g.
+                `envs/myenv.yml`. This file will be used to generate the environment to
+                run in.
+            environment_variables: e.g. `{"PYTHONHASHSEED": "0"}`. These environment
+                variables will be set in the remote environment
+            ssh_key_aws_secret: The name of an AWS secret that contains the contents
+                of a private SSH key that has read access to `repo_url`,
+                e.g. `"my_ssh_key"`. See How to use a private git
+                repo for [AWS](/how_to/private_git_repo_aws.md) or
+                [Azure](/how_to/private_git_repo_azure.md)
+            ssh_key_azure_secret: The name of a secret in the meadowrun-generated
+                Azure Vault (which will be named "mr" followed by the last 22
+                letters/numbers of the Azure subscription id).
+
+        Returns:
+            A `Deployment` object that can be passed to the `run_*` functions.
         """
-        if bool(branch) ^ bool(commit):
+        if branch and commit:
             raise ValueError("Only one of branch and commit can be specified")
 
         if not branch and not commit:
@@ -138,10 +186,10 @@ class Deployment:
             code: Union[CodeDeployment, VersionedCodeDeployment, None] = GitRepoBranch(
                 repo_url=repo_url,
                 branch=branch,
-                # the generated protobuf types say that path_in_repo must not be None,
+                # the generated protobuf types say that path_to_source must not be None,
                 # but this is not actually the case, it handles None fine, which we want
                 # to take advantage of
-                path_in_repo=cast(str, path_in_repo),
+                path_to_source=cast(str, path_to_source),
             )
         else:  # commit
             # guaranteed by prior ifs, this is just for mypy
@@ -150,18 +198,24 @@ class Deployment:
                 repo_url=repo_url,
                 commit=commit,
                 # see comment above
-                path_in_repo=cast(str, path_in_repo),
+                path_to_source=cast(str, path_to_source),
             )
 
         if conda_yml_file:
             interpreter: Optional[InterpreterDeployment] = EnvironmentSpecInCode(
-                environment_type=EnvironmentSpecInCode.EnvironmentType.CONDA,
+                environment_type=EnvironmentType.CONDA,
                 path_to_spec=conda_yml_file,
             )
         else:
             interpreter = None
 
-        if ssh_key_aws_secret:
+        if ssh_key_aws_secret is not None and ssh_key_azure_secret is not None:
+            # TODO this doesn't seem totally right--we could theoretically specify both
+            # if meadowrun was smart enough to use the right one depending on the
+            # environment
+            raise ValueError("Cannot specify both an AWS secret and an Azure secret")
+
+        if ssh_key_aws_secret is not None:
             credentials_sources = [
                 CredentialsSourceForService(
                     service="GIT",
@@ -172,6 +226,18 @@ class Deployment:
                     ),
                 )
             ]
+        elif ssh_key_azure_secret is not None:
+            credentials_sources = [
+                CredentialsSourceForService(
+                    service="GIT",
+                    service_url=repo_url,
+                    source=AzureSecret(
+                        credentials_type=Credentials.Type.SSH_KEY,
+                        vault_name=get_meadowrun_vault_name(get_subscription_id_sync()),
+                        secret_name=ssh_key_azure_secret,
+                    ),
+                )
+            ]
         else:
             credentials_sources = []
 
@@ -187,6 +253,8 @@ def _credentials_source_message(
     )
     if isinstance(credentials_source.source, AwsSecret):
         result.aws_secret.CopyFrom(credentials_source.source)
+    elif isinstance(credentials_source.source, AzureSecret):
+        result.azure_secret.CopyFrom(credentials_source.source)
     elif isinstance(credentials_source.source, ServerAvailableFile):
         result.server_available_file.CopyFrom(credentials_source.source)
     else:
@@ -236,249 +304,126 @@ def _add_defaults_to_deployment(
     )
 
 
-@dataclasses.dataclass
-class JobCompletion(Generic[_T]):
-    """Information about how a job completed"""
+class _DeploymentTarget(Enum):
+    LOCAL = "local"
+    S3 = "s3"
+    AZURE_BLOB_STORAGE = "azure blob storage"
 
-    # TODO both JobCompletion and MeadowrunException should be revisited
-
-    result: _T
-    process_state: ProcessState._ProcessStateEnum.ValueType
-    log_file_name: str
-    return_code: int
-
-
-class MeadowrunException(Exception):
-    def __init__(self, process_state: ProcessState) -> None:
-        super().__init__("Failure while running a meadowrun job: " + str(process_state))
-        self.process_state = process_state
-
-
-class Host(abc.ABC):
-    @abc.abstractmethod
-    async def run_job(self, job: Job) -> JobCompletion[Any]:
-        pass
-
-
-@dataclasses.dataclass(frozen=True)
-class LocalHost(Host):
-    async def run_job(self, job: Job) -> JobCompletion[Any]:
-        initial_update, continuation = await run_local(job)
-        if (
-            initial_update.state != ProcessState.ProcessStateEnum.RUNNING
-            or continuation is None
-        ):
-            result = initial_update
+    @staticmethod
+    def from_cloud_provider(cloud_provider: CloudProviderType) -> _DeploymentTarget:
+        if cloud_provider == "EC2":
+            return _DeploymentTarget.S3
+        elif cloud_provider == "AzureVM":
+            return _DeploymentTarget.AZURE_BLOB_STORAGE
         else:
-            result = await continuation
+            raise ValueError(f"Unknown cloud provider {cloud_provider}")
 
-        if result.state == ProcessState.ProcessStateEnum.SUCCEEDED:
-            job_spec_type = job.WhichOneof("job_spec")
-            # we must have a result from functions, in other cases we can optionally
-            # have a result
-            if job_spec_type == "py_function" or result.pickled_result:
-                unpickled_result = pickle.loads(result.pickled_result)
-            else:
-                unpickled_result = None
-
-            return JobCompletion(
-                unpickled_result, result.state, result.log_file_name, result.return_code
-            )
+    @staticmethod
+    def from_single_host(host: Host) -> _DeploymentTarget:
+        if isinstance(host, AllocCloudInstance):
+            return _DeploymentTarget.from_cloud_provider(host.cloud_provider)
         else:
-            raise MeadowrunException(result)
+            return _DeploymentTarget.LOCAL
+
+    @staticmethod
+    def from_hosts(hosts: AllocCloudInstances) -> _DeploymentTarget:
+        return _DeploymentTarget.from_cloud_provider(hosts.cloud_provider)
+
+
+async def _prepare_code_deployment(
+    code_deploy: Union[CodeDeployment, VersionedCodeDeployment],
+    target: _DeploymentTarget,
+) -> Union[CodeDeployment, VersionedCodeDeployment]:
+    if not isinstance(code_deploy, CodeZipFile):
+        return code_deploy
+
+    if target == _DeploymentTarget.LOCAL:
+        return code_deploy
+    elif target == _DeploymentTarget.S3:
+        # need to deploy zip file to S3, and update the CodeZipFile
+        file_url = urllib.parse.urlparse(code_deploy.url)
+        if file_url.scheme != "file":
+            raise ValueError(f"Expected file URI: {code_deploy.url}")
+        bucket_name, object_name = await s3.ensure_uploaded(file_url.path)
+        s3_url = urllib.parse.urlunparse(("s3", bucket_name, object_name, "", "", ""))
+        code_deploy.url = s3_url
+        shutil.rmtree(os.path.dirname(file_url.path), ignore_errors=True)
+        return code_deploy
+    elif target == _DeploymentTarget.AZURE_BLOB_STORAGE:
+        raise NotImplementedError("Azure Blob Storage is not implemented yet")
+    else:
+        raise ValueError(f"Unexpected value for target {target}")
 
 
 @dataclasses.dataclass(frozen=True)
-class SshHost(Host):
+class AllocCloudInstance(Host):
     """
-    Tells run_function and related functions to connects to the remote machine over SSH
-    via the fabric library https://www.fabfile.org/ fabric_kwargs are passed directly to
-    fabric.Connection().
+    Specifies the requirements for a cloud instance (e.g. an EC2 instance or Azure VM).
+
+    Attributes:
+        logical_cpu_required:
+        memory_gb_required:
+        interruption_probability_threshold: Specifies what interruption probability
+            percent is acceptable. E.g. `80` means that any instance type with an
+            interruption probability less than 80% can be used. Use `0` to indicate that
+            only on-demand instance are acceptable (i.e. do not use spot instances)
+        cloud_provider: `EC2` or `AzureVM`
+        region_name:
     """
 
-    address: str
-    # these options are forwarded directly to Fabric
-    fabric_kwargs: Optional[Dict[str, Any]] = None
-
-    async def run_job(self, job: Job) -> JobCompletion[Any]:
-        with fabric.Connection(
-            self.address, **(self.fabric_kwargs or {})
-        ) as connection:
-            job_io_prefix = ""
-
-            try:
-                # assumes that meadowrun is installed in /var/meadowrun/env as per
-                # build_meadowrun_amis.md. Also uses the default working_folder, which
-                # should (but doesn't strictly need to) correspond to
-                # agent._set_up_working_folder
-
-                # try the first command 3 times, as this is when we actually try to
-                # connect to the remote machine.
-                home_result = await _retry(
-                    lambda: connection.run("echo $HOME"),
-                    (
-                        cast(Exception, paramiko.ssh_exception.NoValidConnectionsError),
-                        cast(Exception, TimeoutError),
-                    ),
-                )
-                if not home_result.ok:
-                    raise ValueError(
-                        "Error getting home directory on remote machine "
-                        + home_result.stdout
-                    )
-
-                remote_working_folder = f"{home_result.stdout.strip()}/meadowrun"
-                mkdir_result = connection.run(f"mkdir -p {remote_working_folder}/io")
-                if not mkdir_result.ok:
-                    raise ValueError(
-                        "Error creating meadowrun directory " + mkdir_result.stdout
-                    )
-
-                job_io_prefix = f"{remote_working_folder}/io/{job.job_id}"
-
-                # serialize job_to_run and send it to the remote machine
-                with io.BytesIO(job.SerializeToString()) as job_to_run_serialized:
-                    connection.put(
-                        job_to_run_serialized, remote=f"{job_io_prefix}.job_to_run"
-                    )
-
-                # fabric doesn't have any async APIs, which means that in order to run
-                # more than one fabric command at the same time, we need to have a
-                # thread per fabric command. We use an asyncio.Future here to make the
-                # API async, so from the user perspective, it feels like this function
-                # is async
-
-                # fabric is supposedly not threadsafe, but it seems to work as long as
-                # more than one connection is not being opened at the same time:
-                # https://github.com/fabric/fabric/pull/2010/files
-                result_future: asyncio.Future = asyncio.Future()
-                event_loop = asyncio.get_running_loop()
-
-                command = (
-                    f"/var/meadowrun/env/bin/meadowrun-local --job-id {job.job_id} "
-                    # TODO --needs-deallocation should only be passed in if we were
-                    # originally using an EC2AllocHost
-                    f"--working-folder {remote_working_folder} --needs-deallocation"
-                )
-
-                print(f"Running {command}")
-
-                def run_and_wait() -> None:
-                    try:
-                        # use meadowrun to run the job
-                        returned_result = connection.run(command)
-                        event_loop.call_soon_threadsafe(
-                            lambda r=returned_result: result_future.set_result(r)
-                        )
-                    except Exception as e2:
-                        event_loop.call_soon_threadsafe(
-                            lambda e2=e2: result_future.set_exception(e2)
-                        )
-
-                threading.Thread(target=run_and_wait).start()
-
-                result = await result_future
-
-                # TODO consider using result.tail, result.stdout
-
-                # see if we got a normal return code
-                if result.return_code != 0:
-                    raise ValueError(f"Process exited {result.return_code}")
-
-                with io.BytesIO() as result_buffer:
-                    connection.get(f"{job_io_prefix}.process_state", result_buffer)
-                    result_buffer.seek(0)
-                    process_state = ProcessState()
-                    process_state.ParseFromString(result_buffer.read())
-
-                if process_state.state == ProcessState.ProcessStateEnum.SUCCEEDED:
-                    job_spec_type = job.WhichOneof("job_spec")
-                    # we must have a result from functions, in other cases we can
-                    # optionally have a result
-                    if job_spec_type == "py_function" or process_state.pickled_result:
-                        result = pickle.loads(process_state.pickled_result)
-                    else:
-                        result = None
-                    return JobCompletion(
-                        result,
-                        process_state.state,
-                        process_state.log_file_name,
-                        process_state.return_code,
-                    )
-                else:
-                    raise MeadowrunException(process_state)
-            finally:
-                if job_io_prefix:
-                    remote_paths = " ".join(
-                        [
-                            f"{job_io_prefix}.job_to_run",
-                            f"{job_io_prefix}.state",
-                            f"{job_io_prefix}.result",
-                            f"{job_io_prefix}.process_state",
-                            f"{job_io_prefix}.initial_process_state",
-                        ]
-                    )
-                    try:
-                        # -f so that we don't throw an error on files that don't
-                        # exist
-                        connection.run(f"rm -f {remote_paths}")
-                    except Exception as e:
-                        print(
-                            f"Error cleaning up files on remote machine: "
-                            f"{remote_paths} {e}"
-                        )
-
-                    # TODO also clean up log files?
-
-
-@dataclasses.dataclass(frozen=True)
-class EC2AllocHost(Host):
-    """A placeholder for a host that will be allocated/created by ec2_alloc.py"""
-
     logical_cpu_required: int
     memory_gb_required: float
     interruption_probability_threshold: float
+    cloud_provider: CloudProviderType
     region_name: Optional[str] = None
 
     async def run_job(self, job: Job) -> JobCompletion[Any]:
-        region_name = self.region_name or await _get_default_region_name()
-        pkey = ensure_meadowrun_key_pair(region_name)
-
-        hosts = await allocate_ec2_instances(
-            Resources(self.memory_gb_required, self.logical_cpu_required, {}),
-            1,
-            self.interruption_probability_threshold,
-            region_name,
-        )
-
-        fabric_kwargs: Dict[str, Any] = {
-            "user": "ubuntu",
-            "connect_kwargs": {"pkey": pkey},
-        }
-
-        if len(hosts) != 1:
-            raise ValueError(f"Asked for one host, but got back {len(hosts)}")
-        host, job_ids = list(hosts.items())[0]
-        if len(job_ids) != 1:
-            raise ValueError(f"Asked for one job allocation but got {len(job_ids)}")
-
-        # Kind of weird that we're changing the job_id here, but okay as long as job_id
-        # remains mostly an internal concept
-        job.job_id = job_ids[0]
-
-        return await SshHost(host, fabric_kwargs).run_job(job)
+        if self.cloud_provider == "EC2":
+            return await run_job_ec2_instance_registrar(
+                job,
+                self.logical_cpu_required,
+                self.memory_gb_required,
+                self.interruption_probability_threshold,
+                self.region_name,
+            )
+        elif self.cloud_provider == "AzureVM":
+            return await run_job_azure_vm_instance_registrar(
+                job,
+                self.logical_cpu_required,
+                self.memory_gb_required,
+                self.interruption_probability_threshold,
+                self.region_name,
+            )
+        else:
+            raise ValueError(
+                f"Unexpected value for cloud_provider {self.cloud_provider}"
+            )
 
 
 @dataclasses.dataclass(frozen=True)
-class EC2AllocHosts:
+class AllocCloudInstances:
     """
-    A placeholder for a set of hosts that will be allocated/created by ec2_alloc.py
+    Specifies the requirements for a set of cloud instances (e.g. EC2 instances or Azure
+    VMs) for running multiple workers. Also see
+    [AllocCloudInstance][meadowrun.AllocCloudInstance]
+
+    Attributes:
+        logical_cpu_required_per_task:
+        memory_gb_required_per_task:
+        interruption_probability_threshold: See
+            [AllocCloudInstance][meadowrun.AllocCloudInstance]
+        cloud_provider: Either `EC2` or `AzureVM`
+        num_concurrent_tasks: The number of workers to launch. In the context of a
+            [run_map][meadowrun.run_map] call, this can be less than or equal to the
+            number of args/tasks. Will default to half the total number of tasks if set
+            to None.
+        region_name:
     """
 
     logical_cpu_required_per_task: int
     memory_gb_required_per_task: float
     interruption_probability_threshold: float
-    # defaults to half the number of total tasks
+    cloud_provider: CloudProviderType
     num_concurrent_tasks: Optional[int] = None
     region_name: Optional[str] = None
 
@@ -531,12 +476,20 @@ async def run_function(
     """
     Runs function on a remote machine, specified by "host".
 
-    Function can either be a reference to a function, a lambda, or a string like
-    "package.module.function_name" (the last option is useful if the function cannot be
-    referenced in the current environment but can be referenced in the deployed
-    environment)
+    Args:
+        function: A reference to a function (e.g. `package.module.function_name`), a
+            lambda, or a string like `"package.module.function_name"` (which is useful
+            if the function cannot be referenced in the current environment but can be
+            referenced in the deployed environment)
+        host: See [AllocCloudInstance][meadowrun.AllocCloudInstance]. Specifies what
+            resources are needed to run this function
+        deployment: See [Deployment][meadowrun.Deployment]. Specifies the
+            environment (code and libraries) that are needed to run this function
+        args: Passed to the function like `function(*args)`
+        kwargs: Passed to the function like `function(**kwargs)`
 
-    The remote machine must have meadowrun installed as per build_meadowrun_amis.md
+    Returns:
+        The result of calling `function`
     """
 
     pickle_protocol = _pickle_protocol_for_deployed_interpreter()
@@ -591,6 +544,10 @@ async def run_function(
         credentials_sources,
     ) = _add_defaults_to_deployment(deployment)
 
+    code = await _prepare_code_deployment(
+        code, _DeploymentTarget.from_single_host(host)
+    )
+
     job = Job(
         job_id=job_id,
         job_friendly_name=friendly_name,
@@ -601,7 +558,8 @@ async def run_function(
     )
     _add_deployments_to_job(job, code, interpreter)
 
-    return (await host.run_job(job)).result
+    job_completion = await host.run_job(job)
+    return job_completion.result
 
 
 async def run_command(
@@ -611,8 +569,20 @@ async def run_command(
     context_variables: Optional[Dict[str, Any]] = None,
 ) -> JobCompletion[None]:
     """
-    Runs the specified command on a remote machine. See run_function_remote for more
-    details on requirements for the remote host.
+    Runs the specified command on a remote machine
+
+    Args:
+        args: Specifies the command to run, can be a string (e.g. `"jupyter nbconvert
+            --to html analysis.ipynb"`) or a list of strings (e.g. `["jupyter",
+            --"nbconvert", "--to", "html", "analysis.ipynb"]`)
+        host: See [AllocCloudInstance][meadowrun.AllocCloudInstance]. Specifies what
+            resources are needed to run this command
+        deployment: See [Deployment][meadowrun.Deployment]. Specifies
+            the environment (code and libraries) that are needed to run this command
+        context_variables: Experimental feature
+
+    Returns:
+        A JobCompletion object that contains metadata about the running of the job.
     """
 
     job_id = str(uuid.uuid4())
@@ -629,6 +599,10 @@ async def run_command(
         credentials_sources,
     ) = _add_defaults_to_deployment(deployment)
 
+    code = await _prepare_code_deployment(
+        code, _DeploymentTarget.from_single_host(host)
+    )
+
     if context_variables:
         pickled_context_variables = pickle.dumps(
             context_variables, protocol=_pickle_protocol_for_deployed_interpreter()
@@ -654,39 +628,55 @@ async def run_command(
 async def run_map(
     function: Callable[[_T], _U],
     args: Sequence[_T],
-    hosts: EC2AllocHosts,
+    hosts: AllocCloudInstances,
     deployment: Optional[Deployment] = None,
 ) -> Sequence[_U]:
-    """Equivalent to map(function, args), but runs distributed."""
+    """
+    Equivalent to `map(function, args)`, but runs distributed and in parallel.
+
+    Args:
+        function: A reference to a function (e.g. `package.module.function_name`) or a
+            lambda
+        args: A list of objects, each item in the list represents a "task",
+            where each "task" is an invocation of `function` on the item in the list
+        hosts: See [AllocCloudInstances][meadowrun.AllocCloudInstances]. Specifies how
+            many workers to provision and what resources are needed for each worker.
+        deployment: See [Deployment][meadowrun.Deployment]. Specifies the environment
+            (code and libraries) that are needed to run the function
+
+    Returns:
+        Returns the result of running `function` on each of `args`
+    """
 
     if not hosts.num_concurrent_tasks:
         num_concurrent_tasks = len(args) // 2 + 1
     else:
         num_concurrent_tasks = min(hosts.num_concurrent_tasks, len(args))
 
-    region_name = hosts.region_name or await _get_default_region_name()
-    pkey = ensure_meadowrun_key_pair(region_name)
-
-    # the first stage of preparation, which happens concurrently:
-
-    # 1. get hosts
-    allocated_hosts_future = asyncio.create_task(
-        allocate_ec2_instances(
-            Resources(
-                hosts.memory_gb_required_per_task,
-                hosts.logical_cpu_required_per_task,
-                {},
-            ),
-            num_concurrent_tasks,
+    if hosts.cloud_provider == "EC2":
+        helper = await prepare_ec2_run_map(
+            function,
+            args,
+            hosts.region_name,
+            hosts.logical_cpu_required_per_task,
+            hosts.memory_gb_required_per_task,
             hosts.interruption_probability_threshold,
-            region_name,
+            num_concurrent_tasks,
         )
-    )
+    elif hosts.cloud_provider == "AzureVM":
+        helper = await prepare_azure_vm_run_map(
+            function,
+            args,
+            hosts.region_name,
+            hosts.logical_cpu_required_per_task,
+            hosts.memory_gb_required_per_task,
+            hosts.interruption_probability_threshold,
+            num_concurrent_tasks,
+        )
+    else:
+        raise ValueError(f"Unexpected value for cloud_provider {hosts.cloud_provider}")
 
-    # 2. create SQS queues and add tasks to the request queue
-    queues_future = asyncio.create_task(create_queues_and_add_tasks(region_name, args))
-
-    # 3. prepare some variables for constructing the worker jobs
+    # prepare some variables for constructing the worker jobs
     friendly_name = _get_friendly_name(function)
     (
         interpreter,
@@ -694,25 +684,20 @@ async def run_map(
         environment_variables,
         credentials_sources,
     ) = _add_defaults_to_deployment(deployment)
-    pickle_protocol = _pickle_protocol_for_deployed_interpreter()
-    fabric_kwargs: Dict[str, Any] = {"user": "ubuntu", "connect_kwargs": {"pkey": pkey}}
 
-    # now wait for 1 and 2 to complete:
-    request_queue_url, result_queue_url = await queues_future
-    allocated_hosts = await allocated_hosts_future
+    code = await _prepare_code_deployment(code, _DeploymentTarget.from_hosts(hosts))
+
+    pickle_protocol = _pickle_protocol_for_deployed_interpreter()
 
     # Now we will run worker_loop jobs on the hosts we got:
 
     pickled_worker_function = cloudpickle.dumps(
-        functools.partial(
-            worker_loop, function, request_queue_url, result_queue_url, region_name
-        ),
-        protocol=pickle_protocol,
+        helper.worker_function, protocol=pickle_protocol
     )
 
     worker_tasks = []
     worker_id = 0
-    for public_address, worker_job_ids in allocated_hosts.items():
+    for public_address, worker_job_ids in helper.allocated_hosts.items():
         for worker_job_id in worker_job_ids:
             job = Job(
                 job_id=worker_job_id,
@@ -730,16 +715,23 @@ async def run_map(
             _add_deployments_to_job(job, code, interpreter)
 
             worker_tasks.append(
-                asyncio.create_task(SshHost(public_address, fabric_kwargs).run_job(job))
+                asyncio.create_task(
+                    SshHost(
+                        public_address,
+                        helper.fabric_kwargs,
+                        (hosts.cloud_provider, helper.region_name),
+                    ).run_job(job)
+                )
             )
 
             worker_id += 1
 
     # finally, wait for results:
 
-    results = await get_results(result_queue_url, region_name, len(args))
+    results = await helper.results_future
 
-    # not really necessary except interpreter will complain...
+    # TODO if there's an error these workers will crash before the results_future
+    # returns
     await asyncio.gather(*worker_tasks)
 
     return results
@@ -762,6 +754,8 @@ def _add_deployments_to_job(
         job.git_repo_commit.CopyFrom(code_deployment)
     elif isinstance(code_deployment, GitRepoBranch):
         job.git_repo_branch.CopyFrom(code_deployment)
+    elif isinstance(code_deployment, CodeZipFile):
+        job.code_zip_file.CopyFrom(code_deployment)
     else:
         raise ValueError(f"Unknown code deployment type {type(code_deployment)}")
 
@@ -775,6 +769,8 @@ def _add_deployments_to_job(
         job.container_at_tag.CopyFrom(interpreter_deployment)
     elif isinstance(interpreter_deployment, EnvironmentSpecInCode):
         job.environment_spec_in_code.CopyFrom(interpreter_deployment)
+    elif isinstance(interpreter_deployment, EnvironmentSpec):
+        job.environment_spec.CopyFrom(interpreter_deployment)
     else:
         raise ValueError(
             f"Unknown interpreter deployment type {type(interpreter_deployment)}"
